/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "TextEvent.hh"
#include "DVector.hh"
#include "jsonStream.hh"
#include "fzstream.hh"
#include <fstream>
#include <stdexcept>

using namespace std;

namespace generator {

   //====================================  Default constructor
   TextEvent::TextEvent(void)
      : DataSource(kSingle), mEvtIndex(0)
   {
   }

   //====================================  Data constructor
   TextEvent::TextEvent(const char* files, const char* channel)
      : DataSource(kSingle), mChannel(channel), mEvtIndex(0)
   {
      setFileList(files);
   }

   //====================================  TextEvent source destructor.
   TextEvent::~TextEvent(void) {
   }

   //====================================  Clone this instance.
   TextEvent*
   TextEvent::clone(void) const {
      return new TextEvent(*this);
   }

   //====================================  Get the default sample-rate;
   double
   TextEvent::defaultRate(const Time& t0, Interval dT) {
      double rate = 0;

      if (mEvtIndex != event_vect.size() && event_vect[mEvtIndex]) {
	 rate = double(event_vect[mEvtIndex]->getTStep());
	 if (rate != 0) rate = 1.0/rate;
      }
      return rate;
   }

   //=====================================  Get the default data type
   long
   TextEvent::defaultType(const Time& t0, Interval dT) {
      long dty = -1;
      if (mEvtIndex != event_vect.size() && event_vect[mEvtIndex]) {
	 dty = event_vect[mEvtIndex]->refDVect()->getType();
      } 
      return dty;
   }

   //=====================================  Get the default data type
   Interval 
   TextEvent::getDuration(void) const {
      Interval dt(1.0);
      if (mEvtIndex < event_vect.size()) {
	 dt = event_vect[mEvtIndex]->getInterval();
      }
      if (debug() > 1) {
	 cout << "TextEvent: Event: " << getName() << "[" << mEvtIndex
	      << "] duration is " << double(dt) << endl;
      }
      return dt;
   }

   //=====================================  Get the default data type
   Time 
   TextEvent::nextTime(const Time& current) {
      Time r;
      size_t N = event_vect.size();
      switch (mTMode) {
      case kSynch:
	 if (mEvtIndex != N) r = event_vect[mEvtIndex]->getStartTime();
	 break;
      case kDelay:
	 if (mEvtIndex != N) r = event_vect[mEvtIndex]->getStartTime() + mDelay;
	 break;
      case kOnce:
	 if (mEvtIndex != N) r = DataSource::nextTime(current);
	 break;
      case kWheel:
	 if (mEvtIndex == N) mEvtIndex = 0;
	 r = DataSource::nextTime(current);
      }
      if (debug() > 1) cout << "TextEvent: Get next " << getName()
			    << " current=" << current.totalS()
			    << " nextTime=" << r << endl;
      return r;
   }
   
   //====================================  Read one or more events
   void
   TextEvent::readEvents(std::istream& in) {
      jsonStream js;

      js.parse(in);
      if (js.isArray()) {
	 for (size_t i=0; i<js.arrayLength(); i++) {
	    try {
	       event_vect.push_back(std::move(js.getTSeries(i)));
	       if (!mChannel.empty() && mChannel!=event_vect.back()->getName()){
		  event_vect.pop_back();
	       }
	    } catch (std::exception& e) {
	       cerr << "Exception in TextEvent: " << e.what() << endl;
	    }
	 }
      } else {
	 try {
	    event_vect.push_back(std::move(js.getTSeries(mChannel)));
	 } catch (std::exception& e) {
	    cerr << "Exception in TextEvent: " << e.what() << endl;
	 }
      }
      if (debug()) cout << "textEvent: Source: " << getName() << " read "
			<< event_vect.size() << " events." << endl; 
   }

   //===================================  Set the channel name
   void
   TextEvent::setChannel(const std::string& name) {
      mChannel = name;
   }

   //===================================  Set the delay time
   void
   TextEvent::setDelay(const std::string& delay) {
      mDelay = Interval(stod(delay));
   }

   //===================================  Read in all the requested event files
   void 
   TextEvent::setFileList(const std::string& files) {
      size_t l = files.size();
      const char* path = files.c_str();
      if (files.substr(l-3) == ".js" || files.substr(l-5) == ".json") {
	 ifstream in(path);
	 readEvents(in);
      } else if (files.substr(l-6) == ".js.gz" || files.substr(l-4) == ".jsz") {
	 fz_istream in(path);
	 readEvents(in);
      } else {
	 ifstream list(path);
	 if (list.is_open()) {
	    std::string line;
	    while (getline(list, line)) {
	       line.erase(0, line.find_first_not_of(" \t"));
	       if (line[0] == '#') continue;
	       line.erase(line.find_last_not_of(" \t")+1);
	       if (line.substr(l-3) == ".js" ||
		   line.substr(l-5) == ".json") {
		  ifstream in(line.c_str());
		  readEvents(in);
	       } else if (line.substr(l-6) == ".js.gz" ||
			  line.substr(l-4) == ".jsz") {
		  fz_istream zin(line.c_str());
		  readEvents(zin);
	       } else {
		  cerr << "TextEvent: Unrecognized event file: " << line << endl;
	       }
	    }
	 } else {
	    cerr << "TextEvent: Unable to open list file: " << files << endl;
	 }
      }
   }

   //====================================  Set the time mode
   void 
   TextEvent::setMode(const string& value) {
      if (value == "synch") {
	 mTMode = kSynch;
      } else if (value == "wheel") {
	 mTMode = kWheel;
      } else if (value == "delay") {
	 mTMode = kDelay;
      } else if (value == "once") {
	 mTMode = kOnce;
      } else {
	 throw runtime_error(string("TextEvent: invalid mode option: ") 
			     + value);
      }
      DataSource::setParameter("Mode", value);
   }

   //====================================  Set an arbitrary parameter
   void 
   TextEvent::setParameter(const string& name, const string& value) {
      if (getDebug()) cout << "Setting TextEvent::" << name << " to " 
			   << value << endl;
      if (name == "Channel")    setChannel(value);
      else if (name == "Files") setFileList(value);
      else if (name == "Mode")  setMode(value);
      else if (name == "Delay") setDelay(value);
      else                      DataSource::setParameter(name, value);
   }
   
   //================================  Get the event time series
   void
   TextEvent::getSourceTSeries(const Time& t0, Interval dT, int N, 
			       TSeries& data) {
      Interval duration = dT * double(N);
      if (debug()) cout << "TextEvent: getting source: " << getName() 
			<< " data for gps " << t0 << "[" << duration
			<< "]" << endl;

      //-----------------------------  Check that data exists.
      if (mEvtIndex >= event_vect.size()) return;
      if (!event_vect[mEvtIndex] || event_vect[mEvtIndex]->empty()) return;
      
      Time tReq(t0);
      TSeries ts;
      switch (mTMode) {
      case kSynch:
	 ts = event_vect[mEvtIndex]->extract(tReq, duration);
	 break;
      case kDelay:
	 tReq -= Interval(mDelay);
	 ts  = event_vect[mEvtIndex]->extract(tReq, duration);
	 break;
      case kOnce:
      case kWheel:
	 //-----------------------------  Record replay start time
	 if (!mStartSeg) mStartSeg = t0;
	 tReq = event_vect[mEvtIndex]->getStartTime() + (t0 - mStartSeg);
	 ts  = event_vect[mEvtIndex]->extract(tReq, duration);
	 break;
      }

      //---------------------------------  Reset the Time & type
      data = TSeries(t0, dT, *ts.refDVect());
      data.setUnits(event_vect[mEvtIndex]->getUnits());

      //---------------------------------  All done with this event
      if (tReq + duration >= event_vect[mEvtIndex]->getEndTime()) {
	 mEvtIndex++;
	 mStartSeg = Time(0);
      } else if (debug() >1) {
	 cout << "not end at " << t0 << " need "
	      << event_vect[mEvtIndex]->getEndTime() - (tReq + duration) << endl;
	 cout << "event_vector: " << endl;
	 event_vect[mEvtIndex]->dump_header(cout) << endl;
	 cout << "return vector: " << endl;
	 data.dump_header(cout) << endl;
      }
   }
}
